#ifndef ASM_E2K_ASSEMBLER_INL_H
#define ASM_E2K_ASSEMBLER_INL_H

#include "registers.h"
#include "assembler-e2k.h"

namespace asm_e2k {

const int Assembler::_c1f1_shift[] = { 21, -1, 14, 7, -1, 0 };


PredDefMap::PredDefMap() : first_set(SizeInBits - 1), last_set(0) {
    std::fill(data, data + SizeInWords, 0);
}

void PredDefMap::set(uint32_t idx) {
    first_set = std::min(first_set, idx);
    last_set  = std::max(last_set, idx);
    data[word_index(idx)] |= bit_mask(idx);
}

bool PredDefMap::get(uint32_t idx) {
    return data[word_index(idx)] & bit_mask(idx);
}

void PredDefMap::reset() {
    uint32_t first_set_word = word_index(first_set);
    uint32_t last_set_word = word_index(last_set);
    for(uint32_t i = first_set_word; i <= last_set_word; i++) {
        data[i] = 0;
    }
#ifdef ASM_E2K_DEBUG
    for(uint32_t i = 0; i < SizeInWords; i++) {
        ASM_E2K_ASSERT(data[i]==0);

    }
#endif
    first_set = SizeInBits - 1;
    last_set = 0;
}

void PredDefMap::set_pred_def(const uint32_t def_idx) {
    set(def_idx);
}

void PredDefMap::set_pred_def(const uint32_t def_idx, const uint32_t pred_idx, bool inverse) {
    int32_t pr_idx;

    ASM_E2K_ASSERT(pred_idx != EREG_PRX || inverse == false);
    if (pred_idx == EREG_PRX) {
        pr_idx = def_idx;
    } else {
        uint32_t pred_idx_ = pred_idx - EREG_PR0;
        ASM_E2K_ASSERT(pred_idx_ >=0 && pred_idx_ < PredicatesUsedCount);
        pr_idx = TotalRegCount * (pred_idx_ * 2 + inverse + 1) + def_idx;
    }
    ASM_E2K_ASSERT(pr_idx <= SizeInBits);
    set(pr_idx);
}

bool PredDefMap::get_pred_def(const uint32_t def_idx) {
    return get(def_idx);
}

bool PredDefMap::get_pred_def(const uint32_t def_idx, const uint32_t pred_idx, bool inverse) {
    int32_t pr_idx;
    ASM_E2K_ASSERT(pred_idx != EREG_PRX || inverse == false);
    if (pred_idx == EREG_PRX) {
        pr_idx = def_idx;
    } else {
        uint32_t pred_idx_ = pred_idx - EREG_PR0;
        ASM_E2K_ASSERT(pred_idx_ >=0 && pred_idx_ < PredicatesUsedCount);
        pr_idx = TotalRegCount * (pred_idx_ * 2 + inverse + 1) + def_idx;
    }
    return get(pr_idx);
}

void AsmScheduler::init_scheduler(Assembler *a, int inst_cnt) {

    insts_cnt = inst_cnt;
    assm = a;

    defs.clear();
    invaruses.clear();
    temp_regs.clear();
    if (curr_num != 0) {
       insts_cnt = curr_num;
    }
#ifdef ASM_SCHED_DEBUG
    printf("init_scheduler clear %d elements curr_num = %d size = %d\n",
            insts_cnt, curr_num,sizeof (struct SchedInstruction) * insts_cnt );
#endif
    curr_num = 0;
    max_depth = 0;
    last_load = 0;
    last_store = 0;
}

void Assembler::init_asm_sheduler(int level, bool sched_debug, bool need_patch, int free_reg, int free_reg_cnt, int insts_cnt) {

    sched_level = level;
    if (!asm_sched) {
        asm_sched = new (&sched_buf) AsmScheduler;
        insts_cnt = MAX_INSTRUCTS_IN_BLOCK;
#ifdef ASM_SCHED_DEBUG
        printf("new (&sched_buf) AsmScheduler sched_debug:%d &sched_buf: %x\n",
                  sched_debug, &sched_buf);
#endif
    }
    patch_offsets = need_patch;
    if (patch_offsets) {
      rbs_offsets.clear();
    }
    sched = asm_sched;
    sched->debug = sched_debug;
    sched->free_register = free_reg;
    sched->free_registers_cnt = free_reg_cnt;
    sched->init_scheduler(this, insts_cnt);
}

Assembler::Assembler() : bw(this), defs(), sched_level(0)
{
    lts_num = 0;
#ifndef V8_TARGET_ARCH_E2K
    buffer_size_ = 2 * KB;
    buffer_ = new byte[buffer_size_];
    is_external_buffer = false;
    pc_ = buffer_;
#endif
    prev_pc_offset = -1;
    _bundle_state = Close;
}

#ifndef V8_TARGET_ARCH_E2K
Assembler::Assembler(byte* buffer, unsigned offset, unsigned buffer_size) : bw(this), buffer_(buffer),
    buffer_size_(buffer_size), is_external_buffer(true), sched_level(0), defs()
{
    ASM_E2K_ASSERT(offset < buffer_size);
    lts_num = 0;
    pc_ = buffer + offset;
    prev_pc_offset = -1;
    _bundle_state = Close;
}

void Assembler::set_buffer(byte* buffer, unsigned offset, unsigned buffer_size) {
    ASM_E2K_ASSERT(is_external_buffer);
    ASM_E2K_ASSERT(offset < buffer_size);
    buffer_ = buffer;
    pc_ = buffer + offset;
    buffer_size_ = buffer_size;
}
#endif

Assembler::~Assembler()
{
#ifndef V8_TARGET_ARCH_E2K
    if(!is_external_buffer) {
      delete [] buffer_;
    }
#endif
}

} // namespace asm_e2k

#endif // ASM_E2K_ASSEMBLER_INL_H

