/*
 * bitMap.hpp
 *
 *  Created on: Feb 17, 2017
 *      Author: msyabro
 */

#ifndef UTILITIES_BITMAP_HPP_
#define UTILITIES_BITMAP_HPP_

#include "tango/tango_globals.hpp"
#include "tango/tango_allocator.hpp"
#include <cstdint>
#include <algorithm>
#ifdef __e2k__
#include <e2kbuiltin.h>
#endif // __e2k__

// Operations for bitmaps represented as arrays of unsigned integers.
// Bit offsets are numbered from 0 to size-1.

#define USE_FIRST_LAST_SET_IDX 1
#if USE_FIRST_LAST_SET_IDX
#define IF_USE_FIRST_LAST_SET_IDX(a) a
#else
#define IF_USE_FIRST_LAST_SET_IDX(a)
#endif

namespace tango {

class BitMap : public TangoObj {
 public:
  typedef size_t idx_t;         // Type used for bit and word indices.
  typedef intptr_t bm_word_t;  // Element type of array that represents
                                // the bitmap.

  // Hints for range sizes.
  typedef enum {
    unknown_range, small_range, large_range
  } RangeSizeHint;

 private:
  Allocator* _allocator;
  bm_word_t* _map;     // First word in bitmap
  idx_t      _size;    // Size of bitmap (in bits)
  idx_t      _size_in_words;
#if USE_FIRST_LAST_SET_IDX
  idx_t      _first_set; // First non-zero bit;
  idx_t      _last_set;  // Last non-zero bit;
#endif

  // Puts the given value at the given offset, using resize() to size
  // the bitmap appropriately if needed using factor-of-two expansion.
  void at_put_grow(idx_t index, bool value);
  bm_word_t* word_addr(idx_t bit) const { return map() + word_index(bit); }
  static idx_t bit_in_word(idx_t bit) { return bit & (BitsPerWord - 1); }
  static idx_t word_index(idx_t bit)  { return bit >> LogBitsPerWord; }
  static bm_word_t bit_mask(idx_t bit) { return (bm_word_t)1 << bit_in_word(bit); }
  static idx_t bit_index(idx_t word)  { return word << LogBitsPerWord; }

  bm_word_t
  inverted_bit_mask_for_range(idx_t beg, idx_t end) const {
    assert_rvm(end != 0, "does not work when end == 0");
    assert_rvm(beg == end || word_index(beg) == word_index(end - 1),
           "must be a single-word range");
    bm_word_t mask = bit_mask(beg) - 1;   // low (right) bits
    if (bit_in_word(end) != 0) {
      mask |= ~(bit_mask(end) - 1);       // high (left) bits
    }
    return mask;
  }

  idx_t word_index_round_up(idx_t bit) const {
    idx_t bit_rounded_up = bit + (BitsPerWord - 1);
    // Check for integer arithmetic overflow.
    idx_t word_idx = word_index(bit_rounded_up);
    return word_idx <= size_in_words() ? word_idx : size_in_words();
  }

  bm_word_t* map() const           { return _map; }
  bm_word_t  map(idx_t word) const { return _map[word]; }

#ifdef TANGO_ASSERT
  inline void verify_index(idx_t index) const {
    assert_rvm(index < _size, "BitMap index out of bounds");
  }

  inline void verify_range(idx_t beg_index, idx_t end_index) const {
    assert_rvm(beg_index <= end_index, "BitMap range error");
    // Note that [0,0) and [size,size) are both valid ranges.
    if (end_index != _size) verify_index(end_index);
  }
#else
  inline void verify_index(idx_t index) const { }
  inline void verify_range(idx_t beg_index, idx_t end_index) const { }
#endif // #ifdef TANGO_ASSERT

  // Statistics.
  static idx_t* _pop_count_table;
  static void init_pop_count_table();
  static idx_t num_set_bits(bm_word_t w);
  static idx_t num_set_bits_from_table(unsigned char c);

  bool range_intersection(idx_t begin1, idx_t end1, idx_t begin2, idx_t end2, idx_t &out_begin, idx_t &out_end) const {
    out_begin = std::max(begin1, begin2);
    out_end = std::min(end1, end2);
    return out_end >= out_begin;
  }

  void clear_range_of_words(idx_t beg, idx_t end) {
      bm_word_t *map = _map;
      //std::fill(map + beg, map + end, 0);
      for (idx_t i = beg; i < end; ++i) map[i] = 0;
  }

  void clear_large_range_of_words(idx_t beg, idx_t end) {
    memset(_map + beg, 0, (end - beg) * sizeof(uintptr_t));
  }

  void set_range_of_words(idx_t beg, idx_t end) {
    bm_word_t* map = _map;
    for (idx_t i = beg; i < end; ++i) map[i] = ~(uintptr_t)0;
  }

  void set_large_range_of_words(idx_t beg, idx_t end) {
    memset(_map + beg, ~(unsigned char)0, (end - beg) * sizeof(uintptr_t));
  }

 public:

  // Constructs a bitmap with no map, and size 0.
  BitMap() : _allocator(NULL), _map(NULL), _size(0), _size_in_words(0)
#if USE_FIRST_LAST_SET_IDX
             ,_first_set(-1), _last_set(0)
#endif
  { }

  // Constructs a bitmap with the given map and size.
  BitMap(bm_word_t* map, idx_t size_in_bits);

  // Constructs an empty bitmap of the given size (that is, this clears the
  // new bitmap).  Allocates the map array in resource area if
  // "in_resource_area" is true, else in the C heap.
  BitMap(idx_t size_in_bits, bool in_resource_area = true);

  // Set the map and size.
  void set_map(bm_word_t* map)      { _map = map; }
  void set_size(idx_t size_in_bits) { _size = size_in_bits; _size_in_words = (_size  + BitsPerWord - 1) >> LogBitsPerWord; }

  // Allocates necessary data structure, either in the resource area
  // or in the C heap, as indicated by "in_resource_area."
  // Preserves state currently in bit map by copying data.
  // Zeros any newly-addressable bits.
  // If "in_resource_area" is false, frees the current map.
  // (Note that this assumes that all calls to "resize" on the same BitMap
  // use the same value for "in_resource_area".)
  void resize(idx_t size_in_bits, bool in_resource_area = true);

  // Accessing
  idx_t size() const                    { return _size; }
//  idx_t size_in_words() const           { return _size >> LogBitsPerWord + BitsPerWord - 1; }
  idx_t size_in_words() const           { return _size_in_words; }

  bool at(idx_t index) const {
    verify_index(index);
    return (*word_addr(index) & bit_mask(index)) != 0;
  }

  // Align bit index up or down to the next bitmap word boundary, or check
  // alignment.
  static idx_t word_align_up(idx_t bit) {
    return align_size_up(bit, BitsPerWord);
  }
  static idx_t word_align_down(idx_t bit) {
    return align_size_down(bit, BitsPerWord);
  }
  static bool is_word_aligned(idx_t bit) {
    return word_align_up(bit) == bit;
  }

  // Set or clear the specified bit.
  inline void set_bit(idx_t bit) {
    verify_index(bit);
    *word_addr(bit) |= bit_mask(bit);
#if USE_FIRST_LAST_SET_IDX
    _first_set = std::min(_first_set, bit);
    _last_set  = std::max(_last_set, bit);
#endif
  }
  inline void clear_bit(idx_t bit) {
    verify_index(bit);
    *word_addr(bit) &= ~bit_mask(bit);
  }

  // Atomically set or clear the specified bit.
  inline bool par_set_bit(idx_t bit) {
    return false;
  }
  inline bool par_clear_bit(idx_t bit) {
    return false;
  }

  // Put the given value at the given offset. The parallel version
  // will CAS the value into the bitmap and is quite a bit slower.
  // The parallel version also returns a value indicating if the
  // calling thread was the one that changed the value of the bit.
  void at_put(idx_t index, bool value);
  bool par_at_put(idx_t index, bool value);

  // Update a range of bits.  Ranges are half-open [beg, end).
  void set_range   (idx_t beg, idx_t end);
  void clear_range (idx_t beg, idx_t end);
  void set_large_range   (idx_t beg, idx_t end);
  void clear_large_range (idx_t beg, idx_t end);
  void at_put_range(idx_t beg, idx_t end, bool value);
  void par_at_put_range(idx_t beg, idx_t end, bool value);
  void at_put_large_range(idx_t beg, idx_t end, bool value);
  void par_at_put_large_range(idx_t beg, idx_t end, bool value);

  void  clear_range_within_word    (idx_t beg, idx_t end);
  void  set_range_within_word      (idx_t beg, idx_t end);

  // Update a range of bits, using a hint about the size.  Currently only
  // inlines the predominant case of a 1-bit range.  Works best when hint is a
  // compile-time constant.
  void set_range(idx_t beg, idx_t end, RangeSizeHint hint);
  void clear_range(idx_t beg, idx_t end, RangeSizeHint hint);
  void par_set_range(idx_t beg, idx_t end, RangeSizeHint hint);
  void par_clear_range  (idx_t beg, idx_t end, RangeSizeHint hint);

  // Clearing
  void clear_large();
  inline void clear() {
    IF_USE_FIRST_LAST_SET_IDX(if(_first_set <= _last_set)) clear_range_of_words(0, size_in_words());
#if USE_FIRST_LAST_SET_IDX
    _first_set = 0;
    _last_set = 0;
#endif
  }
  inline void set_all() {
    set_range_of_words(0, size_in_words());
#if USE_FIRST_LAST_SET_IDX
    _first_set = 0;
    _last_set = (size() == 0) ? 0 : (size() - 1);
#endif
  }


  // Looking for 1's and 0's at indices equal to or greater than "l_index",
  // stopping if none has been found before "r_index", and returning
  // "r_index" (which must be at most "size") in that case.
  idx_t get_next_one_offset_inline (idx_t l_offset, idx_t r_offset) const {
    assert_rvm(l_offset <= size(), "BitMap index out of bounds");
    assert_rvm(r_offset <= size(), "BitMap index out of bounds");
    assert_rvm(l_offset <= r_offset, "l_offset > r_offset ?");

    if (l_offset == r_offset) {
      return l_offset;
    }
#if USE_FIRST_LAST_SET_IDX
    idx_t   index;
    idx_t r_index;
    if(!range_intersection(_first_set, _last_set, l_offset, r_offset, index, r_index)) {
      return r_offset;
    }
    index = word_index(index);
    r_index = word_index(r_index) + 1;
#else
    idx_t   index = word_index(l_offset);
    idx_t r_index = word_index(r_offset-1) + 1;
#endif
    idx_t res_offset = l_offset;
#if USE_FIRST_LAST_SET_IDX
    res_offset = std::max(res_offset, _first_set);
#endif

    // check bits including and to the _left_ of offset's position
    idx_t pos = bit_in_word(res_offset);
    idx_t res = map(index) >> pos;
    if (res != (uintptr_t)NoBits) {
      // find the position of the 1-bit
#ifdef __e2k__
      res_offset += __builtin_e2k_lzcntd(__builtin_e2k_bitrevd(res));
#else
      for (; !(res & 1); res_offset++) {
        res = res >> 1;
      }
#endif // __e2k__

#ifdef TANGO_ASSERT
      // In the following assert, if r_offset is not bitamp word aligned,
      // checking that res_offset is strictly less than r_offset is too
      // strong and will trip the assert.
      //
      // Consider the case where l_offset is bit 15 and r_offset is bit 17
      // of the same map word, and where bits [15:16:17:18] == [00:00:00:01].
      // All the bits in the range [l_offset:r_offset) are 0.
      // The loop that calculates res_offset, above, would yield the offset
      // of bit 18 because it's in the same map word as l_offset and there
      // is a set bit in that map word above l_offset (i.e. res != NoBits).
      //
      // In this case, however, we can assert is that res_offset is strictly
      // less than size() since we know that there is at least one set bit
      // at an offset above, but in the same map word as, r_offset.
      // Otherwise, if r_offset is word aligned then it will not be in the
      // same map word as l_offset (unless it equals l_offset). So either
      // there won't be a set bit between l_offset and the end of it's map
      // word (i.e. res == NoBits), or res_offset will be less than r_offset.

      idx_t limit = is_word_aligned(r_offset) ? r_offset : size();
      assert_rvm(res_offset >= l_offset && res_offset < limit, "just checking");
#endif // TANGO_ASSERT
      return std::min(res_offset, r_offset);
    }
    // skip over all word length 0-bit runs
    for (index++; index < r_index; index++) {
      res = map(index);
      if (res != (uintptr_t)NoBits) {
        // found a 1, return the offset
#ifdef __e2k__
        res_offset = bit_index(index) + __builtin_e2k_lzcntd(__builtin_e2k_bitrevd(res));
#ifdef TANGO_ASSERT
        res = res >> (res_offset - bit_index(index));
#endif // TANGO_ASSERT
#else
        for (res_offset = bit_index(index); !(res & 1); res_offset++) {
          res = res >> 1;
        }
#endif // __e2k__
        assert_rvm(res & 1, "tautology; see loop condition");
        assert_rvm(res_offset >= l_offset, "just checking");
        return std::min(res_offset, r_offset);
      }
    }
    return r_offset;
  }
  idx_t get_next_zero_offset_inline(idx_t l_offset, idx_t r_offset) const {
    assert_rvm(l_offset <= size(), "BitMap index out of bounds");
    assert_rvm(r_offset <= size(), "BitMap index out of bounds");
    assert_rvm(l_offset <= r_offset, "l_offset > r_offset ?");

    if (l_offset == r_offset) {
      return l_offset;
    }
    idx_t   index = word_index(l_offset);
    idx_t r_index = word_index(r_offset-1) + 1;
    idx_t res_offset = l_offset;

    // check bits including and to the _left_ of offset's position
    idx_t pos = res_offset & (BitsPerWord - 1);
    idx_t res = (map(index) >> pos) | left_n_bits((int)pos);

    if (res != (uintptr_t)AllBits) {
      // find the position of the 0-bit
      for (; res & 1; res_offset++) {
        res = res >> 1;
      }
      assert_rvm(res_offset >= l_offset, "just checking");
      return std::min(res_offset, r_offset);
    }
    // skip over all word length 1-bit runs
    for (index++; index < r_index; index++) {
      res = map(index);
      if (res != (uintptr_t)AllBits) {
        // found a 0, return the offset
        for (res_offset = index << LogBitsPerWord; res & 1;
             res_offset++) {
          res = res >> 1;
        }
        assert_rvm(!(res & 1), "tautology; see loop condition");
        assert_rvm(res_offset >= l_offset, "just checking");
        return std::min(res_offset, r_offset);
      }
    }
    return r_offset;
  }

  // Like "get_next_one_offset_inline", except requires that "r_index" is
  // aligned to bitsizeof(bm_word_t).
  idx_t get_next_one_offset_inline_aligned_right(idx_t l_offset,
                                                        idx_t r_offset) const {
    verify_range(l_offset, r_offset);
    assert_rvm(bit_in_word(r_offset) == 0, "r_offset not word-aligned");

    if (l_offset == r_offset) {
      return l_offset;
    }
    idx_t   index = word_index(l_offset);
    idx_t r_index = word_index(r_offset);
    idx_t res_offset = l_offset;

    // check bits including and to the _left_ of offset's position
    idx_t res = map(index) >> bit_in_word(res_offset);
    if (res != (uintptr_t)NoBits) {
      // find the position of the 1-bit
      for (; !(res & 1); res_offset++) {
        res = res >> 1;
      }
      assert_rvm(res_offset >= l_offset &&
             res_offset < r_offset, "just checking");
      return res_offset;
    }
    // skip over all word length 0-bit runs
    for (index++; index < r_index; index++) {
      res = map(index);
      if (res != (uintptr_t)NoBits) {
        // found a 1, return the offset
        for (res_offset = bit_index(index); !(res & 1); res_offset++) {
          res = res >> 1;
        }
        assert_rvm(res & 1, "tautology; see loop condition");
        assert_rvm(res_offset >= l_offset && res_offset < r_offset, "just checking");
        return res_offset;
      }
    }
    return r_offset;
  }

  // Non-inline versionsof the above.
  idx_t get_next_one_offset (idx_t l_index, idx_t r_index) const {
    return get_next_one_offset_inline(l_index, r_index);
  }
  idx_t get_next_zero_offset(idx_t l_index, idx_t r_index) const {
    return get_next_zero_offset_inline(l_index, r_index);
  }

  idx_t get_next_one_offset(idx_t offset) const {
    return get_next_one_offset(offset, size());
  }
  idx_t get_next_zero_offset(idx_t offset) const {
    return get_next_zero_offset(offset, size());
  }

  // Returns the number of bits set in the bitmap.
  idx_t count_one_bits() const;

  // Set operations.
  void set_union(const BitMap& bits);
  void set_difference(const BitMap& bits);
  void set_intersection(const BitMap& bits);
  // Returns true iff "this" is a superset of "bits".
  bool contains(const BitMap& bits) const;
  // Returns true iff "this and "bits" have a non-empty intersection.
  bool intersects(const BitMap& bits) const;

  // Returns result of whether this map changed
  // during the operation
  bool set_union_with_result(const BitMap& bits);
  bool set_difference_with_result(const BitMap& bits);
  bool set_intersection_with_result(const BitMap& bits);

  // Requires the submap of "bits" starting at offset to be at least as
  // large as "this".  Modifies "this" to be the intersection of its
  // current contents and the submap of "bits" starting at "offset" of the
  // same length as "this."
  // (For expedience, currently requires the offset to be aligned to the
  // bitsize of a uintptr_t.  This should go away in the future though it
  // will probably remain a good case to optimize.)
  void set_intersection_at_offset(const BitMap& bits, idx_t offset);

  void set_from(const BitMap& bits);

  bool is_same(const BitMap& bits) const;

  // Test if all bits are set or cleared
  bool is_full() const;
  bool is_empty() const;

};

} // namespace tango

#endif /* UTILITIES_BITMAP_HPP_ */
