#ifndef TANGO_CODE_BUFFER_HPP
#define TANGO_CODE_BUFFER_HPP

#include "tango_globals.hpp"
#include "tango_allocator.hpp"

namespace tango {

class CodeSection : public TangoObj {
private:
  address _buf;
  address _pos;    // Pointer to the first free byte
  address _limit;  // Pointer to the last byte available for writing

public:
  CodeSection() : _buf(nullptr), _pos(nullptr), _limit(nullptr) {}

  CodeSection(address buf, address limit) {
    initialize(buf, limit);
  }

  CodeSection(Allocator* allocator, int32_t size) {
    _buf = reinterpret_cast<address>(allocator->allocate(size));
    _pos = _buf;
    _limit = _buf + size;
  }

  void initialize(address buf, address limit) {
    _buf = buf;
    _limit = limit;
    _pos = buf;
  }

  address pos()         const      { return _pos; }
  address start()       const      { return _buf; }
  address limit()       const      { return _limit; }
  int32_t capacity()    const      { return _limit - _buf; }
  int32_t size()        const      { return _pos - _buf; }
  void advance(uint32_t byte_size) { _pos += byte_size; }

  void check_capacity(int32_t bytes) {
    // TODO: reallocating mechanics
    assert_rvm(_pos + bytes <= _limit, "not enuogh space to allocate code");
  }
};

//----------------------------------------------------------------------------//
class CodeBuffer : public TangoObj {
public:
  const int BUFFER_SIZE = 4 * 1024 * 1024; // 4mb

private:
  CodeSection* _insts; // Section for main code, this section is mandatory
  CodeSection* _async; // Section for async programs
  CodeSection* _cur_section;

public:
  void emit_int8(int8_t x) {
    _cur_section->check_capacity(sizeof(int8_t));
    address pc = _cur_section->pos();
    *((int8_t*) pc) = x;
    _cur_section->advance(sizeof(int8_t));
  }

  void emit_int16(int16_t x) {
    _cur_section->check_capacity(sizeof(int16_t));
    address pc = _cur_section->pos();
    *((int16_t*) pc) = x;
    _cur_section->advance(sizeof(int16_t));
  }

  void emit_int32(int32_t x) {
    _cur_section->check_capacity(sizeof(int32_t));
    address pc = _cur_section->pos();
    *((int32_t*) pc) = x;
    _cur_section->advance(sizeof(int32_t));
  }

  void emit_int64(int64_t x) {
    _cur_section->check_capacity(sizeof(int64_t));
    address pc = _cur_section->pos();
    *((int64_t*) pc) = x;
    _cur_section->advance(sizeof(int64_t));
  }

  void emit_fapb(int64_t value, bool need_align);

  void skip(int32_t bytes) {
    _cur_section->advance(bytes);
  }

  address pc() const {
    return _cur_section->pos();
  }

  address insts_begin() const {
    return _insts->start();
  }

  address async_begin() const {
    return _async->start();
  }

  CodeSection* section() const {
    return _cur_section;
  }

  CodeSection* insts() const {
    return _insts;
  }

  CodeSection* async() const {
    return _async;
  }

  int32_t get_code_len() const {
    return (((((_insts->pos() - _insts->start()) - 1) / 4) + 1) * 4);
  }

  void set_async(address buf, address limit) {
    _async = new CodeSection(buf, limit);
  }

  void set_code_section(CodeSection* cs) {
    _cur_section = cs;
  }

  void copy(address start_inst, size_t size) {
    _cur_section->check_capacity(size);
    memcpy(pc(), start_inst, size);
    _cur_section->advance(size);
  }

  CodeBuffer(address buf, address limit) {
    _insts = new CodeSection(buf, limit);
    _async = nullptr;
    _cur_section = _insts;
  }

  CodeBuffer(Allocator* allocator) {
    _insts = new CodeSection(allocator, BUFFER_SIZE);
    _async = nullptr;
    _cur_section = _insts;
  }
};
 
} // namespace tango

#endif // TANGO_CODE_BUFFER_HPP
