#ifndef TANGO_CPUINFO_HPP
#define TANGO_CPUINFO_HPP

#include "tango_inst.hpp"

namespace tango {

enum class ElbrusModel {
  RESERVED   = 0,
  Elbrus     = 1,
  Elbrus_S   = 2,
  Elbrus_4C  = 3,
  Elbrus_2CP = 4,
  Elbrus_4S  = 5,
  Elbrus_2CM = 6,
  Elbrus_8C  = 7,
  Elbrus_1C  = 8,
  Elbrus_8C2 = 9,
  Elbrus_12C = 10,
  Elbrus_16C = 11,
  Elbrus_2C3 = 12
};

enum class ElbrusISA {
  na = 0,
  v1 = 1,
  v2 = 2,
  v3 = 3,
  v4 = 4,
  v5 = 5,
  v6 = 6
};

union IDR {
  uint64_t value;
  struct {
    uint32_t model      : 8;  // Model number
    uint32_t revision   : 4;  // Revision of the current model
    uint64_t ms         : 52; // Model specific information, implementation dependent
  } bits;
};

class CPUInfo : public TangoObj {
public:
  CPUInfo() {
    fillCPUInfo();
  }

  ElbrusModel getCPUModel() {
    return (ElbrusModel)(idr.bits.model);
  }

  ElbrusISA getCPUArchVersion() {
    return modelToVer((ElbrusModel)idr.bits.model);
  }

  // Model Elbrus corresponds to the v1 version of architecture.
  // IDR register is not available in v1, so we can never
  // encounter Elbrus model here
  ElbrusISA modelToVer(const ElbrusModel& model) {
    switch(model) {
      case ElbrusModel::Elbrus:       TANGO_UNREACHABLE;
                                      return ElbrusISA::na;
      case ElbrusModel::Elbrus_S:
      case ElbrusModel::Elbrus_2CP:
      case ElbrusModel::Elbrus_2CM:   return ElbrusISA::v2;
      case ElbrusModel::Elbrus_4C:
      case ElbrusModel::Elbrus_4S:    return ElbrusISA::v3;
      case ElbrusModel::Elbrus_8C:
      case ElbrusModel::Elbrus_1C:    return ElbrusISA::v4;
      case ElbrusModel::Elbrus_8C2:   return ElbrusISA::v5;
      case ElbrusModel::Elbrus_12C:
      case ElbrusModel::Elbrus_16C:
      case ElbrusModel::Elbrus_2C3:   return ElbrusISA::v6;
      default:                        assert_rvm(false, "Unknown architecture");
                                      return ElbrusISA::na;
    }
  }

  void fillCPUInfo() {
    __asm volatile ("rrd %%idr, %0" : "=r"(idr.value));
  }

private:
  IDR idr;
};


} // namespace tango

#endif // TANGO_CPUINFO_HPP
